#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest,time

import logging
from ibm_watson_machine_learning.tests.Cloud.preparation_and_cleaning import *

class TestRshinyApp(unittest.TestCase):
    runtime_uid = None
    deployment_uid = None
    function_uid = None
    scoring_url = None
    logger = logging.getLogger(__name__)
    shiny_asset_id = None

    @classmethod
    def setUpClass(self):
        TestRshinyApp.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.cos_credentials = get_cos_credentials()
        self.cos_resource_crn = self.cos_credentials['resource_instance_id']
        self.instance_crn = get_instance_crn()
        self.space_name = str(uuid.uuid4())

        metadata = {
                     self.client.spaces.ConfigurationMetaNames.NAME: 'space' + self.space_name,
                     self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description',
                     self.client.spaces.ConfigurationMetaNames.STORAGE: {
                                                                           "type": "bmcos_object_storage",
                                                                           "resource_crn": self.cos_resource_crn
                     },
            self.client.spaces.ConfigurationMetaNames.COMPUTE: {
                "name": "existing_instance_id",
                "crn": self.instance_crn
            }
        }

        self.space = self.client.spaces.store(meta_props=metadata, background_mode=False)

        TestRshinyApp.space_id = self.client.spaces.get_id(self.space)
        # TestRshinyApp.space_id = '9cf73498-72d6-42eb-a6f4-424616de9f45'
        print('space_id: ', TestRshinyApp.space_id)
        self.client.set.default_space(TestRshinyApp.space_id)

    # def test_01_service_instance_details(self):
    #     TestAIFunction.logger.info("Check client ...")
    #     self.assertTrue(self.client.__class__.__name__ == 'APIClient')
    #
    #     TestAIFunction.logger.info("Getting instance details ...")
    #     details = self.client.service_instance.get_details()
    #     TestAIFunction.logger.debug(details)
    #
    #     self.assertTrue("published_models" in str(details))
    #     self.assertEqual(type(details), dict)

    def test_01_create_shiny_asset(self):

        self.client.shiny.ConfigurationMetaNames.show()


        meta_prop_shiny = {
            self.client.shiny.ConfigurationMetaNames.NAME: "my shiny app",
            self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "shiny app for deployment"
        }

        shiny_details = self.client.shiny.store(meta_prop_shiny, file_path="artifacts/app.R.zip")

        print(shiny_details)

        TestRshinyApp.shiny_asset_id = self.client.shiny.get_id(shiny_details)
        TestRshinyApp.shiny_asset_url = self.client.shiny.get_href(shiny_details)
        print("shiny asset ID:" + str(TestRshinyApp.shiny_asset_id))
        print("shiny asset URL:" + str(TestRshinyApp.shiny_asset_url))
        self.assertIsNotNone(TestRshinyApp.shiny_asset_id)
        self.assertIsNotNone(TestRshinyApp.shiny_asset_url)

    def test_02_download_shiny_content(self):
        try:
            os.remove('test_shiny_asset.zip')
        except:
            pass
        self.client.shiny.download(TestRshinyApp.shiny_asset_id, filename='test_shiny_asset.zip')
        try:
            os.remove('test_shiny_asset.zip')
        except:
            pass

    def test_03_get_details(self):

        details = self.client.shiny.get_details(TestRshinyApp.shiny_asset_id)
        print(details)
        self.assertTrue(TestRshinyApp.shiny_asset_id in str(details))

    def test_04_list(self):
        self.client.shiny.list()

    def test_05_revisions(self):
        meta_prop_shiny_revision1 = {
            self.client.shiny.ConfigurationMetaNames.NAME: "Shiny revision 1",
            self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "Shiny revision 1"
        }

        # Update meta and attachment for revision 1 creation
        shiny_update = self.client.shiny.update(TestRshinyApp.shiny_asset_id,
                                                meta_prop_shiny_revision1,
                                                file_path="artifacts/app1.R.zip")

        print('update: ', shiny_update)

        attachment_id = shiny_update[u'metadata'][u'attachment_id']

        # Create revision
        revision = self.client.shiny.create_revision(TestRshinyApp.shiny_asset_id)

        print('revision: ', revision)

        self.assertTrue(revision[u'metadata'][u'revision_id'] and revision[u'metadata'][u'revision_id'] == 1)

        new_attachment_id = revision[u'metadata'][u'attachment_id']

        self.assertTrue(attachment_id != new_attachment_id)

        # List revisions
        self.client.shiny.list_revisions(TestRshinyApp.shiny_asset_id)

        self.client.shiny.get_revision_details(TestRshinyApp.shiny_asset_id, 1)

        try:
            os.remove('test_shiny_asset_rev1.zip')
        except:
            pass

        # Download revision 1 attachment
        self.client.shiny.download(TestRshinyApp.shiny_asset_id, filename='test_shiny_asset_rev1.zip', rev_uid=1)
        try:
            os.remove('test_shiny_asset_rev1.zip')
        except:
            pass

    def test_06_create_deployment(self):
        deploy_meta = {
                self.client.deployments.ConfigurationMetaNames.NAME: "deployment_rshiny",
                self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "deployment rshiny deployment",
                self.client.deployments.ConfigurationMetaNames.R_SHINY: {"authentication" : "anyone_with_url" },
                self.client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {"name":"S", "num_nodes":1}
            }

        TestRshinyApp.logger.info("Create deployment")
        deployment = self.client.deployments.create(artifact_uid=TestRshinyApp.shiny_asset_id, meta_props=deploy_meta)
        TestRshinyApp.logger.debug("deployment: " + str(deployment))
        # TestRshinyApp.scoring_url = self.client.deployments.get_scoring_href(deployment)
        # TestRshinyApp.logger.debug("Scoring href: {}".format(TestRshinyApp.scoring_url))
        TestRshinyApp.deployment_uid = self.client.deployments.get_uid(deployment)
        TestRshinyApp.logger.debug("Deployment uid: {}".format(TestRshinyApp.deployment_uid))
        self.client.deployments.list()
        self.assertTrue("deployment_rshiny" in str(deployment))

    def test_07_update_deployment(self):
        patch_meta = {
            self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "deployment_Updated_Shiny_Description",
        }
        self.client.deployments.update(TestRshinyApp.deployment_uid, patch_meta)

    def test_08_get_deployment_details(self):
        TestRshinyApp.logger.info("Get deployment details")
        deployment_details = self.client.deployments.get_details()
        print(deployment_details)
        TestRshinyApp.logger.debug("Deployment details: {}".format(deployment_details))
        self.assertTrue('deployment_Updated_Shiny_Description' in str(deployment_details))

    def test_09_delete_deployment(self):
        TestRshinyApp.logger.info("Delete deployment")
        self.client.deployments.delete(TestRshinyApp.deployment_uid)

    def test_10_delete_shiny_asset(self):
        TestRshinyApp.logger.info("Delete function")
        self.client.shiny.delete(TestRshinyApp.shiny_asset_id)
        self.client.spaces.delete(TestRshinyApp.space_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import os
from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location, FSLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d, create_connection_to_cos
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestOBM

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, RegressionAlgorithms


class TestAutoAIRemote(AbstractTestOBM, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """
    HISTORICAL_RUNS_CHECK = False
    cos_resource = None

    input_data_filenames = ['group_customer_main.csv', 'group_customer_customers.csv',
                            'group_customer_transactions.csv',
                            'group_customer_purchase.csv', 'group_customer_products.csv']
    input_data_path = './autoai/data/'

    input_node_names = [name.replace('group_customer_', '').split('.')[0] for name in input_data_filenames]

    cos_endpoint = "https://s3.us.cloud-object-storage.appdomain.cloud"
    data_cos_path = "data/"

    SPACE_ONLY = False

    BATCH_DEPLOYMENT_WITH_DA = False
    BATCH_DEPLOYMENT_WITH_CDA = False
    BATCH_DEPLOYMENT_WITH_CA_DA = True
    BATCH_DEPLOYMENT_WITH_CA_CA = True

    OPTIMIZER_NAME = "Group Customer OBM test sdk"

    target_space_id: str = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='test description',
        prediction_type=PredictionType.REGRESSION,
        prediction_column='next_purchase',
        scoring=Metrics.MEDIAN_ABSOLUTE_ERROR,
        holdout_size=0.2,
        include_only_estimators=[RegressionAlgorithms.LGBM, RegressionAlgorithms.XGB]
    )

    def test_00b_prepare_COS_instance_and_connection(self):
        TestAutoAIRemote.connection_id, TestAutoAIRemote.bucket_name = create_connection_to_cos(
            wml_client=self.wml_client,
            cos_credentials=self.cos_credentials,
            cos_endpoint=self.cos_endpoint,
            bucket_name=self.bucket_name,
            save_data=True,
            data_path=self.input_data_path,
            data_filenames=self.input_data_filenames,
            data_cos_path=self.data_cos_path)

        self.assertIsInstance(self.connection_id, str)

    def test_01_create_multiple_data_connections__connections_created(self):
        TestAutoAIRemote.data_connections = []
        for file, node_name in zip(self.input_data_filenames, self.input_node_names):
            conn = DataConnection(
                data_join_node_name=node_name,
                connection_asset_id=self.connection_id,
                location=S3Location(
                    bucket=self.bucket_name,
                    path=os.path.join(self.data_cos_path, file)
                )
            )
            TestAutoAIRemote.data_connections.append(conn)

        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connections)
        self.assertGreater(len(TestAutoAIRemote.data_connections), 0)
        self.assertIsNone(self.results_connection)

    def test_02_create_data_join_graph__graph_created(self):
        data_join_graph = DataJoinGraph()
        data_join_graph.node(name="main")
        data_join_graph.node(name="customers")
        data_join_graph.node(name="transactions")
        data_join_graph.node(name="purchase")
        data_join_graph.node(name="products")
        data_join_graph.edge(from_node="main", to_node="customers",
                             from_column=["group_customer_id"], to_column=["group_customer_id"])
        data_join_graph.edge(from_node="main", to_node="transactions",
                             from_column=["transaction_id"], to_column=["transaction_id"])
        data_join_graph.edge(from_node="main", to_node="purchase",
                             from_column=["group_id"], to_column=["group_id"])
        data_join_graph.edge(from_node="transactions", to_node="products",
                             from_column=["product_id"], to_column=["product_id"])

        TestAutoAIRemote.data_join_graph = data_join_graph

        print(f"data_join_graph: {data_join_graph}")

    # def test_08_fit_run_training_of_auto_ai_in_wml(self):
    #
    #     AbstractTestOBM.remote_auto_pipelines = self.experiment.runs.get_optimizer("d7ed9b03-0577-42d4-938c-14ff5f1ad528")
    #     AbstractTestOBM.trained_pipeline_details = self.remote_auto_pipelines.get_run_details()
    #     AbstractTestOBM.run_id = self.trained_pipeline_details['metadata']['id']
    #
    #     status = self.remote_auto_pipelines.get_run_status()
    #     run_details = self.remote_auto_pipelines.get_run_details().get('entity')
    #     self.assertNotIn(status, ['failed', 'canceled'], msg=f"Training finished with status {status}. \n"
    #                                                          f"Details: {run_details.get('status')}")


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

class GenericArchiveFrameworkCheck(object):
    @staticmethod
    def is_archive_framework(name):
        if (re.match('spss-modeler', name) is not None) or (re.match('pmml', name) is not None) \
          or (re.match('caffe', name) is not None) or (re.match('caffe2', name) is not None) \
          or (re.match('torch',name) is not None) \
          or (re.match('pytorch', name) is not None) or (re.match('blueconnect', name) is not None) \
          or (re.match('mxnet', name) is not None) or (re.match('theano', name) is not None) \
          or (re.match('darknet', name) is not None):
          return True
        else:
         return False



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import mlpipelinepy


class MLPipelineVersion(type):
    _version = mlpipelinepy.__version__
    _version_parts = _version.split(".")

    @classmethod
    def major_ver(mcs):
        return int(mcs._version_parts[0])

    @classmethod
    def minor_ver(mcs):
        return int(mcs._version_parts[1])

    @classmethod
    def full_version(mcs):
        return mcs._version



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .model_collection import ModelCollection
from .experiment_collection import ExperimentCollection
from .wml_experiment_collection import WmlExperimentCollection

from .function_collection import FunctionCollection
from .libraries_collection import LibrariesCollection
from ibm_watson_machine_learning.libs.repo.ml_api_client import MLApiClient
from ibm_watson_machine_learning.libs.repo.ml_authorization import MLAuthorization
from ibm_watson_machine_learning.libs.repo.mlrepository.meta_props import MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepositoryclient.ml_repository_api import MLRepositoryApi
from .runtimes_collection import RuntimesCollection
import sys
if sys.version_info < (3, 0):
    from urlparse import urlparse
else:
    from urllib.parse import urlparse


class MLRepositoryClient(MLAuthorization):
    """
    Main class used as client to connect to Repository sevice.

    :param str base_path: base url to Watson Machine Learning instance
    :param MLApiClient api_client: customized api_client provided by user

    :ivar ModelCollection models: provides client to operate on models

    """
    def __init__(self, base_path=None, api_client=None):

        if base_path is not None:
            if api_client is None:
                api_client = MLApiClient(base_path)

            if not isinstance(base_path, str) and not isinstance(base_path, unicode):
                raise ValueError('Invalid type for base_path: {}'.format(base_path.__class__.__name__))

            if not isinstance(api_client, MLApiClient):
                raise ValueError('Invalid type for api_client: {}'.format(api_client.__class__.__name__))

            self.repository_api = MLRepositoryApi(api_client)
            super(MLRepositoryClient, self).__init__(api_client)

            self.models = ModelCollection(base_path, self.repository_api, self)
            self.pipelines = ExperimentCollection(base_path, self.repository_api, self)
            self.experiments = WmlExperimentCollection(base_path, self.repository_api, self)
            self.functions = FunctionCollection(base_path, self.repository_api, self)
            self.libraries = LibrariesCollection(base_path, self.repository_api, self)
            self.runtimes = RuntimesCollection(base_path, self.repository_api, self)

        else:
            self.repository_api = None
            super(MLRepositoryClient, self).__init__(api_client)
            self.models = None
            self.pipelines = None
            self.experiments = None
            self.functions = None


    def iam_connect(self, wml_vcap):

        if not isinstance(wml_vcap, dict):
            raise TypeError("Expecting object of type : %s" % dict+ " but got %s" % type(wml_vcap))

        base_path = wml_vcap.get('url')
        if base_path is None:
            raise TypeError("Watson ML service credentials: url not defined")

        parsed = urlparse(base_path)
        if parsed.scheme == '' or parsed.netloc == '':
            raise TypeError("Watson ML service credentials: Invalid URL")
        api_client = MLApiClient(base_path)
        self.repository_api = MLRepositoryApi(api_client)
        self.api_client = api_client
        self.models = ModelCollection(base_path, self.repository_api, self)
        self.pipelines = ExperimentCollection(base_path, self.repository_api, self)
        self.experiments = WmlExperimentCollection(base_path, self.repository_api, self)
        self.functions = FunctionCollection(base_path, self.repository_api, self)

        self.iam_authorize(wml_vcap)


    @staticmethod
    def meta(self):
        """returns meta object

    >>> model_artifact = MLRepositoryArtifact(model, name='test-model-name', training_data=training, meta_props=MetaProps({
    >>> MetaNames.EVALUATION_METRICS: json.dumps([{
    >>>   "name": "accuracy",
    >>>    "value": 0.95,
    >>>    "threshold": 0.9}]) }))
    >>> created_model_artifact = ml_repository_client.models.save(model_artifact)
    >>> created_model_metaprops = created_model_artifact.meta.prop(MetaNames.EVALUATION_METRICS)

        """
        return MetaProps({})

def connect(vcap):
    """
    Authorizes user to connect by providing the vcap

   :param dict vcap: WML Credentials with API key or WML credentials

   """

    if not isinstance(vcap, dict):
        raise TypeError("Expecting object of type : %s" % dict+ " but got %s" % type(vcap))
    iam_client = MLRepositoryClient()
    iam_client.iam_connect(vcap)
    return iam_client



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import ibm_boto3
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, ContainerLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d, save_data_to_container
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestWebservice, \
    AbstractTestAutoAIAsync, AbstractTestBatch

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, ClassificationAlgorithms


class TestAutoAIRemote(AbstractTestAutoAIAsync, AbstractTestWebservice, AbstractTestBatch, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/Corona_NLP_test_utf8.csv'

    data_cos_path = 'data/Corona_NLP_test_utf8.csv'

    SPACE_ONLY = False

    OPTIMIZER_NAME = "Corona NLP Text Transformer test sdk"

    batch_payload_location = './autoai/data/Corona_NLP_scoring_payload.csv'
    batch_payload_cos_location = "Corona_NLP_scoring_payload.csv"

    BATCH_DEPLOYMENT_WITH_CA = True
    BATCH_DEPLOYMENT_WITH_CDA = False
    BATCH_DEPLOYMENT_WITH_DA = False
    BATCH_DEPLOYMENT_WITH_DF = True

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        prediction_type=PredictionType.MULTICLASS,
        prediction_column='Sentiment',
        scoring=Metrics.F1_SCORE_MACRO,
        holdout_size=0.1,
        text_processing=True,
        text_columns_names=['OriginalTweet', 'Location'],
        word2vec_feature_number=4,
        max_number_of_estimators=1,
        daub_give_priority_to_runtime=3.0,
    )

    def test_00b_write_data_to_container(self):
        if self.SPACE_ONLY:
            self.wml_client.set.default_space(self.space_id)
        else:
            self.wml_client.set.default_project(self.project_id)

        save_data_to_container(self.data_location, self.data_cos_path, self.wml_client)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(
            location=ContainerLocation(path=self.data_cos_path
                                       ))
        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)

    def test_08_get_run_details(self):
        parameters = self.remote_auto_pipelines.get_run_details()
        print(parameters)
        self.assertIsNotNone(parameters)

    def test_08a_get_feature_importance(self):
        feature_importance_df = self.remote_auto_pipelines.get_pipeline_details(pipeline_name='Pipeline_1').get('features_importance')
        self.assertIsNotNone(feature_importance_df)

        str_feature_importance_index_list = str(list(feature_importance_df.sort_index().index))

        text_columns = self.experiment_info['text_columns_names']

        for column in text_columns:
            self.assertIn(f'word2vec({column})', str_feature_importance_index_list,
                          msg=f"word2vec({column}) is not in features importance table. Full table: {feature_importance_df}")

        self.assertIn('NewTextFeature_0', str_feature_importance_index_list,
                      msg="Text features were numerated incorrectly.")
        self.assertIn(f"NewTextFeature_{self.experiment_info['word2vec_feature_number']*len(text_columns)-1}",
                      str_feature_importance_index_list, msg="Text features were numerated incorrectly.")


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import requests
import json
import base64
import logging
from datetime import datetime, timedelta
from ibm_watson_machine_learning.wml_client_error import NoWMLCredentialsProvided, ApiRequestFailure, WMLClientError
from ibm_watson_machine_learning.href_definitions import HrefDefinitions


class ServiceInstance:
    """
        Connect, get details and check usage of your Watson Machine Learning service instance.
    """
    def __init__(self, client):
        self._logger = logging.getLogger(__name__)
        self._client = client
        self._ICP = client.ICP
        self._wml_credentials = client.wml_credentials
        if self._ICP:
            if self.get_instance_id() == 'openshift' or self.get_instance_id() == 'wml_local':
                self._wml_credentials[u'url'] = self.get_url()
            else:
                self._wml_credentials[u'url'] = self.get_url() + ':31843'
            self._wml_credentials[u'instance_id'] = "999"
        self._href_definitions = HrefDefinitions(self._wml_credentials)
        self._client.wml_token = self._create_token()
        self._logger.info(u'Successfully prepared token: ' + self._client.wml_token)
        # ml_repository_client is initialized in repo
        self.details = None

    def get_instance_id(self):
        """
             Get instance id of your Watson Machine Learning service.
             
             **Output**
             
             .. important::
             
                **returns**: instance id\n
                **return type**: str
             
             **Example**
             
             >>> instance_details = client.service_instance.get_instance_id()
        """
        return self._wml_credentials['instance_id']

    def get_api_key(self):
        """
            Get api_key  of Watson Machine Learning service.

            **Output**

            .. important::
                **returns** api_key\n
                **return type**: str\n

            **Example**

                >>> instance_details = client.service_instance.get_api_key()
        """
        return self._wml_credentials['apikey']

    def get_url(self):
        """
             Get instance url of your Watson Machine Learning service.
             
             **Output**
             
             .. important::
             
                **returns**: instance url\n
                **return type**: str

             **Example**

             >>> instance_details = client.service_instance.get_url()
        """
        return self._wml_credentials['url']

    def get_username(self):
        """
             Get username for your Watson Machine Learning service.
             
             **Output**
             
             .. important::
             
                **returns**: username\n
                **return type**: str
             
             **Example**
             
             >>> instance_details = client.service_instance.get_username()
        """
        return self._wml_credentials['username']

    def get_password(self):
        """
             Get password for your Watson Machine Learning service.
             
             **Output**
             
             .. important::
             
                **returns**: password\n
                **return type**: str

             **Example**

             >>> instance_details = client.service_instance.get_password()
        """
        return self._wml_credentials['password']

    def get_details(self):
        """
             Get information about your Watson Machine Learning instance.
             
             **Output**
            
             .. important::

                **returns**: metadata of service instance\n
                **return type**: dict

             **Example**

             >>> instance_details = client.service_instance.get_details()
        """
        if not self._ICP:
            if self._wml_credentials is not None:

                # if self._client.CREATED_IN_V1_PLAN:
                #     response_get_instance = requests.get(
                #         self._href_definitions.get_v4_instance_id_href(),
                #         # params=self._client._params(),
                #         params={'version': self._client.version_param},
                #         headers=self._client._get_headers()
                #     )
                # else:
                response_get_instance = requests.get(
                    self._href_definitions.get_instance_by_id_endpoint_href(),
                    headers=self._client._get_headers()
                )

                if response_get_instance.status_code == 200:
                    return response_get_instance.json()
                else:
                    raise ApiRequestFailure(u'Getting instance details failed.', response_get_instance)
            else:
                raise NoWMLCredentialsProvided
        else:
            return {}

    def _get_token(self):
        if self._client.wml_token is None:
            self._create_token()
            self._client.repository._refresh_repo_client()

        else:
            if self._client._is_IAM():
                if self._get_expiration_datetime() - timedelta(minutes=15) < datetime.now():
                    self._client.wml_token = self._get_IAM_token()
                    self._client.repository._refresh_repo_client()

            elif self._get_expiration_datetime() - timedelta(minutes=30) < datetime.now():
                self._client.repository._refresh_repo_client()
                self._refresh_token()

        return self._client.wml_token

    def _create_token(self):

        if self._client.proceed is True:
            return self._wml_credentials["token"]
        if self._client._is_IAM():
            return self._get_IAM_token()
        else:
            if not self._ICP:
                response = requests.get(self._href_definitions.get_token_endpoint_href(),
                                        auth=(self.get_username(), self.get_password()))
                if response.status_code == 200:
                    token = response.json().get(u'token')
                else:
                    raise ApiRequestFailure(u'Error during getting ML Token.', response)
                return token
            else:
                token_url = self._wml_credentials['url'].replace(':31002', ':31843') + '/v1/preauth/validateAuth'
                response = requests.get(token_url, auth=(self.get_username(), self.get_password()), verify=False)

                if response.status_code == 200:
                    token = response.json().get(u'accessToken')
                else:
                    raise ApiRequestFailure(u'Error during getting ML Token.', response)
                return token

    def _refresh_token(self):
        import requests
        if self._client.proceed is True:
            self._client.wml_token = self._wml_credentials["token"]
        else:
            if not self._ICP:
                response = requests.put(
                    self._href_definitions.get_token_endpoint_href(),
                    json={'token': self._client.wml_token},
                    headers={
                        'Content-Type': 'application/json',
                        'Accept': 'application/json',
                        'X-WML-User-Client': 'PythonClient'
                    }
                )

                if response.status_code == 200:
                    self._client.wml_token = response.json().get(u'token')
                else:
                    raise ApiRequestFailure(u'Error during refreshing ML Token.', response)
            else:
                token_url = self._wml_credentials['url'].replace(':31002', ':31843') + '/v1/preauth/validateAuth'
                response = requests.get(token_url, auth=(self.get_username(), self.get_password()), verify=False)

                if response.status_code == 200:
                    self._client.wml_token = response.json().get(u'accessToken')
                else:
                    raise ApiRequestFailure(u'Error during refreshing ICP Token.', response)

    def _get_expiration_datetime(self):
        token_parts = self._client.wml_token.split('.')
        token_padded = token_parts[1] + '=' * (len(token_parts[1]) % 4)
        token_info = json.loads(base64.b64decode(token_padded).decode('utf-8'))
        token_expire = token_info.get('exp')

        return datetime.fromtimestamp(token_expire)

    def _is_iam(self):
        token_parts = self._client.wml_token.split('.')
        token_padded = token_parts[1] + '=' * (len(token_parts[1]) % 4)
        token_info = json.loads(base64.b64decode(token_padded).decode('utf-8'))
        instanceId = token_info.get('instanceId')

        return instanceId

    def _get_IAM_token(self):
        if self._client.proceed is True:
            return self._wml_credentials["token"]
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Basic Yng6Yng='
        }

        mystr = 'apikey=' + self._href_definitions.get_iam_token_api()
        response = requests.post(
            self._href_definitions.get_iam_token_url(),
            data=mystr,
            headers=headers
        )

        if response.status_code == 200:
            token = response.json().get(u'access_token')
        else:
            raise ApiRequestFailure(u'Error getting IAM Token.', response)
        return token

    def _create_zen_token(self):
        token_url = self._wml_credentials['url'] + '/icp4d-api/v1/authorize'
        headers = {
            'Content-type' :'application/json',
            'cache-control': 'no-cache'
        }
        response = requests.post(token_url, auth=(self.get_username(), self.get_password()), verify=False)
        if response.status_code == 200:
            print(response)
            token = response.json().get(u'token')
        else:
            raise ApiRequestFailure(u'Error during refreshing ML Token.', response)
        return token



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from configparser import ConfigParser
import json
import os
import io
import sys
import ibm_boto3
from ibm_botocore.client import Config
import wget
import uuid


if "ENV" in os.environ:
    environment = os.environ['ENV']
else:
    environment = "SVT"


config = ConfigParser()
config.read('../config.ini')

MNIST = 'mnist'
MNIST_LMDB = 'mnist_lmdb'
BOSTON = 'boston'
BAIR_BVLC = 'bair_bvlc_caffenet'


def get_env():
    return environment


def get_wml_credentials():
    return json.loads(config.get(environment, 'wml_credentials'))


def get_cos_credentials():
    return json.loads(config.get(environment, 'cos_credentials'))


def get_feedback_data_reference():
    return json.loads(config.get(environment, 'feedback_data_reference'))


def get_spark_reference():
    return json.loads(config.get(environment, 'spark_reference'))


def get_cos_auth_endpoint():
    return config.get(environment, 'cos_auth_endpoint')


def get_cos_service_endpoint():
    return config.get(environment, 'cos_service_endpoint')


def get_client():
    wml_lib = __import__('ibm_watson_machine_learning', globals(), locals())
    return wml_lib.APIClient(get_wml_credentials())


def get_cos_resource():
    cos_credentials = get_cos_credentials()
    api_key = cos_credentials['apikey']
    service_instance_id = cos_credentials['resource_instance_id']
    auth_endpoint = get_cos_auth_endpoint()
    service_endpoint = get_cos_service_endpoint()

    cos = ibm_boto3.resource(
        's3',
        ibm_api_key_id = api_key,
        ibm_service_instance_id = service_instance_id,
        ibm_auth_endpoint = auth_endpoint,
        config = Config(signature_version='oauth'),
        endpoint_url = service_endpoint
    )

    return cos


def prepare_cos(cos_resource, bucket_prefix='wml-test', data_code = MNIST):
    client = get_client()
    clean_env(client, cos_resource)
    import datetime

    postfix = datetime.datetime.now().isoformat().replace(":", "-").split(".")[0].replace("T", "-")

    bucket_names = {
        'data': '{}-{}-data-{}'.format(bucket_prefix, environment.lower(), postfix),
        'results': '{}-{}-results-{}'.format(bucket_prefix, environment.lower(), postfix)
    }

    cos_resource.create_bucket(Bucket=bucket_names['data'])
    upload_data(cos_resource, bucket_names['data'], data_code)

    cos_resource.create_bucket(Bucket=bucket_names['results'])

    return bucket_names


def download_data(data_code, data_dir):
    if data_code == MNIST:
        data_links = [
            'http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz',
            'http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz',
            'http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz',
            'http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz',
            'https://s3.amazonaws.com/img-datasets/mnist.npz'
        ]
    elif data_code == BOSTON:
        raise Exception('Unsupported.')
    elif data_code == BAIR_BVLC:
        data_links = [
            'http://dl.caffe.berkeleyvision.org/bvlc_reference_caffenet.caffemodel'
        ]
    else:
        raise Exception('Unrecognized data_code.')

    if data_dir is not None:
        if not os.path.isdir(data_dir):
            os.mkdir(data_dir)

    for link in data_links:
        if not os.path.isfile(os.path.join(data_dir, os.path.join(link.split('/')[-1]))):
            wget.download(link, out=data_dir)


def upload_data(cos_resource, bucket_name, data_code=MNIST):
    if data_code == MNIST:
        data_dir = os.path.join('datasets', 'MNIST_DATA')
    elif data_code == BAIR_BVLC:
        data_dir = os.path.join('datasets', 'BAIR_BVLC')
    elif data_code == BOSTON:
        data_dir = os.path.join('datasets', 'boston', 'BOSTON_DATA')
    elif data_code == MNIST_LMDB:
        data_dir = os.path.join('datasets', 'mnist_data_lmdb', 'mnist_train_lmdb')
    else:
        raise Exception('Unrecognized data_code.')

    try:
        download_data(data_code, data_dir)
    except:
        pass

    bucket_obj = cos_resource.Bucket(bucket_name)

    if data_code == MNIST or data_code == BAIR_BVLC or data_code == BOSTON:
        for filename in os.listdir(data_dir):
            with open(os.path.join(data_dir, filename), 'rb') as data:
                bucket_obj.upload_file(os.path.join(data_dir, filename), filename)
                print('{} is uploaded.'.format(filename))
    elif data_code == MNIST_LMDB:
        bucket_obj.upload_file(os.path.join('datasets', 'mnist_data_lmdb', 'mnist_test_lmdb', 'data.mdb'), 'mnist_test_lmdb/data.mdb')
        bucket_obj.upload_file(os.path.join('datasets', 'mnist_data_lmdb', 'mnist_test_lmdb', 'lock.mdb'), 'mnist_test_lmdb/lock.mdb')
        bucket_obj.upload_file(os.path.join('datasets', 'mnist_data_lmdb', 'mnist_train_lmdb', 'data.mdb'), 'mnist_train_lmdb/data.mdb')
        bucket_obj.upload_file(os.path.join('datasets', 'mnist_data_lmdb', 'mnist_train_lmdb', 'lock.mdb'), 'mnist_train_lmdb/lock.mdb')
    else:
        raise Exception('No such dataset available')


    for obj in bucket_obj.objects.all():
        print('Object key: {}'.format(obj.key))
        print('Object size (kb): {}'.format(obj.size/1024))


def get_cos_training_data_reference(bucket_names):
    cos_credentials = get_cos_credentials()
    service_endpoint = get_cos_service_endpoint()

    return {
        "connection": {
            "endpoint_url": service_endpoint,
            "access_key_id": cos_credentials['cos_hmac_keys']['access_key_id'],
            "secret_access_key": cos_credentials['cos_hmac_keys']['secret_access_key']
        },
        "source": {
            "bucket": bucket_names['data'],
        },
        "type": "s3"
    }


def get_cos_training_results_reference(bucket_names):
    cos_credentials = get_cos_credentials()
    service_endpoint = get_cos_service_endpoint()

    return {
        "connection": {
            "endpoint_url": service_endpoint,
            "access_key_id": cos_credentials['cos_hmac_keys']['access_key_id'],
            "secret_access_key": cos_credentials['cos_hmac_keys']['secret_access_key']
        },
        "target": {
            "bucket": bucket_names['results'],
        },
        "type": "s3"
    }

def clean_cos_bucket(cos_resource, bucket_name):
    bucket_obj = cos_resource.Bucket(bucket_name)
    for upload in bucket_obj.multipart_uploads.all():
        upload.abort()
    for o in bucket_obj.objects.all():
        o.delete()
    bucket_obj.delete()


def clean_cos(cos_resource, bucket_names):
    clean_cos_bucket(cos_resource, bucket_names['data'])
    clean_cos_bucket(cos_resource, bucket_names['results'])


from datetime import datetime, timedelta
from pytz import timezone
yesterday = datetime.now(timezone('UTC')) - timedelta(days=1)


def clean_env(client, cos_resource, threashold_date=yesterday):
    clean_experiments(client, threashold_date)
    clean_training_runs(client, threashold_date)
    clean_definitions(client, threashold_date)
    clean_models(client, threashold_date)
    clean_deployments(client, threashold_date)
    try:
        clean_ai_functions(client, threashold_date)
        clean_runtimes(client, threashold_date)
        clean_custom_libraries(client, threashold_date)
    except:
        pass

    if cos_resource is not None:
        for bucket in cos_resource.buckets.all():
            if 'wml-test-' in bucket.name and bucket.creation_date < threashold_date:
                print('Deleting \'{}\' bucket.'.format(bucket.name))
                try:
                    for upload in bucket.multipart_uploads.all():
                        upload.abort()
                    for o in bucket.objects.all():
                        o.delete()
                    bucket.delete()
                except Exception as e:
                    print("Exception during bucket deletion occured: " + str(e))


def clean_models(client, threashold_date=yesterday):
    details = client.repository.get_model_details()

    for model_details in details['resources']:
        if model_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' model.'.format(model_details['metadata']['guid']))
            try:
                client.repository.delete(model_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of model failed: ' + str(e))


def clean_definitions(client, threashold_date=yesterday):
    details = client.repository.get_definition_details()

    for definition_details in details['resources']:
        if definition_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' definition.'.format(definition_details['metadata']['guid']))
            try:
                client.repository.delete(definition_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of definition failed: ' + str(e))


def clean_deployments(client, threashold_date=yesterday):
    details = client.deployments.get_details()

    for deployment_details in details['resources']:
        if deployment_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' deployment.'.format(deployment_details['metadata']['guid']))
            try:
                client.deployments.delete(deployment_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of deployment failed: ' + str(e))


def clean_experiments(client, threashold_date=yesterday):
    details = client.repository.get_experiment_details()

    for experiment_details in details['resources']:
        if experiment_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' experiment.'.format(experiment_details['metadata']['guid']))
            try:
                client.repository.delete(experiment_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of experiment failed: ' + str(e))


def clean_training_runs(client, threashold_date=yesterday):
    details = client.training.get_details()

    for run_details in details['resources']:
        if run_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' training run.'.format(run_details['metadata']['guid']))
            try:
                client.training.delete(run_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of training run failed: ' + str(e))


def clean_runtimes(client, threashold_date=yesterday):
    details = client.runtimes.get_details()

    for runtime_details in details['resources']:
        if runtime_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' runtime.'.format(runtime_details['metadata']['guid']))
            try:
                client.runtimes.delete(runtime_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of runtime failed: ' + str(e))


def clean_custom_libraries(client, threashold_date=yesterday):
    details = client.runtimes.get_library_details()

    for lib_details in details['resources']:
        if lib_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' custom library.'.format(lib_details['metadata']['guid']))
            try:
                client.runtimes.delete_library(lib_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of custom library failed: ' + str(e))


def clean_ai_functions(client, threashold_date=yesterday):
    details = client.repository.get_function_details()

    for func_details in details['resources']:
        if func_details['metadata']['created_at'] < threashold_date.isoformat():
            print('Deleting \'{}\' AI function.'.format(func_details['metadata']['guid']))
            try:
                client.repository.delete(func_details['metadata']['guid'])
            except Exception as e:
                print('Deletion of AI function failed: ' + str(e))


def run_monitor(client, experiment_run_uid, queue):
    stdout_ = sys.stdout
    captured_output = io.StringIO()  # Create StringIO object
    sys.stdout = captured_output  # and redirect stdout.
    client.experiments.monitor_logs(experiment_run_uid)
    sys.stdout = stdout_  # Reset redirect.

    print(captured_output.getvalue())

    queue.put(captured_output.getvalue())


def run_monitor_metrics(client, experiment_run_uid, queue):
    stdout_ = sys.stdout
    captured_output = io.StringIO()  # Create StringIO object
    sys.stdout = captured_output  # and redirect stdout.
    client.experiments.monitor_metrics(experiment_run_uid)
    sys.stdout = stdout_  # Reset redirect.

    print(captured_output.getvalue())

    queue.put(captured_output.getvalue())


def run_monitor_training(client, training_run_uid, queue):
    stdout_ = sys.stdout
    captured_output = io.StringIO()  # Create StringIO object
    sys.stdout = captured_output  # and redirect stdout.
    client.training.monitor_logs(training_run_uid)
    sys.stdout = stdout_  # Reset redirect.

    print(captured_output.getvalue())

    queue.put(captured_output.getvalue())



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc

import pandas as pd
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, DatabaseLocation
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes.abstract_test_iris_wml_autoai_multiclass_connections import (
    AbstractTestAutoAIRemote)
from ibm_watson_machine_learning.tests.utils import get_db_credentials
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError


class AbstractTestAutoAIDatabaseConnection(AbstractTestAutoAIRemote, abc.ABC):
    """
    The test can be run on CLOUD, and CPD
    The test covers:
    - DataBase connection set-up.
    - Saving data to a database.
    - downloading training data using created connection.
    - downloading all generated pipelines to lale pipeline.
    - deployment with lale pipeline.
    - deployment deletion.
    Connection used in test:
    - input: Connection to database.
    """
    database_name = "db_name"
    schema_name = "tests_sdk"
    table_name = "iris"

    def test_00c_prepare_connection_to_DATABASE(self):
        AbstractTestAutoAIDatabaseConnection.db_credentials = get_db_credentials(self.database_name)
        connection_details = self.wml_client.connections.create({
            'datasource_type': self.wml_client.connections.get_datasource_type_uid_by_name(self.database_name),
            'name': 'Connection to DB for python API tests',
            'properties': self.db_credentials
        })

        AbstractTestAutoAIDatabaseConnection.connection_id = self.wml_client.connections.get_uid(connection_details)
        self.assertIsInstance(self.connection_id, str)

    def test_01_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        AbstractTestAutoAIDatabaseConnection.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                                                 project_id=self.project_id,
                                                                 space_id=self.space_id)

        self.assertIsInstance(self.experiment, AutoAI, msg="Experiment is not of type AutoAI.")

    def test_02_DataConnection_setup(self):
        AbstractTestAutoAIDatabaseConnection.data_connection = DataConnection(
            connection_asset_id=self.connection_id,
            location=DatabaseLocation(
                schema_name=self.schema_name,
                table_name=self.table_name
            )
        )
        AbstractTestAutoAIDatabaseConnection.results_connection = None

        self.assertIsNotNone(obj=AbstractTestAutoAIDatabaseConnection.data_connection)

        self.data_connection._wml_client = self.wml_client
        self.data_connection.write(pd.read_csv(self.data_location))

    def test_02a_read_saved_remote_data_before_fit(self):
        AbstractTestAutoAIDatabaseConnection.data = self.data_connection.read()
        print("Data sample:")
        print(self.data.head())
        self.assertGreater(len(self.data), 0)

    def test_29_delete_connection_and_connected_data_asset(self):
        self.wml_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.connections.get_details(self.connection_id)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import ibm_boto3
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, ContainerLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d, save_data_to_container
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestWebservice, \
    AbstractTestAutoAIAsync

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, RegressionAlgorithms


@unittest.skipIf(is_cp4d(), "Excel files not supported on CPD 4.0 an low")
class TestAutoAIRemote(AbstractTestAutoAIAsync, AbstractTestWebservice, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/xlsx/credit_risk_insurance__two_sheets.xlsx'
    sheet_name = 'insurance'
    sheet_number = 1

    data_cos_path = 'data/credit_risk_insurance__two_sheets.xlsx'

    SPACE_ONLY = True

    OPTIMIZER_NAME = "Insurance test sdk"

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        prediction_type=PredictionType.REGRESSION,
        prediction_column='charges',
        scoring=Metrics.R2_SCORE,
        include_only_estimators=[RegressionAlgorithms.LGBM, RegressionAlgorithms.XGB],
        max_number_of_estimators=2,
        excel_sheet=sheet_name
    )

    def test_00b_write_data_to_container(self):
        if self.SPACE_ONLY:
            self.wml_client.set.default_space(self.space_id)
        else:
            self.wml_client.set.default_project(self.project_id)

        save_data_to_container(self.data_location, self.data_cos_path, self.wml_client)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(
            location=ContainerLocation(path=self.data_cos_path
                                       ))
        TestAutoAIRemote.results_connection = DataConnection(
            location=ContainerLocation(path=self.results_cos_path
                                       ))

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNotNone(obj=TestAutoAIRemote.results_connection)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import os

# SPARK_HOME_PATH = os.environ['SPARK_HOME']
# PYSPARK_PATH = str(SPARK_HOME_PATH) + "/python/"
# sys.path.insert(1, path_join(PYSPARK_PATH))

import logging
from preparation_and_cleaning import *
from models_preparation import *


class TestWMLClientWithSpark(unittest.TestCase):
    model_uid = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithSpark.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.model_name = "SparkMLlibFromObjectLocal Model"
        self.deployment_name = "Test deployment"
        self.project_id = os.environ['PROJECT_ID']
        self.client.set.default_project(self.project_id)

    def test_03_publish_model(self):
        TestWMLClientWithSpark.logger.info("Creating spark model ...")

        model_data = create_spark_mllib_model_data()

        TestWMLClientWithSpark.logger.info("Publishing spark model ...")

        self.client.repository.ModelMetaNames.show()

        model_props = {self.client.repository.ModelMetaNames.NAME: "Spark",
            self.client.repository.ModelMetaNames.TYPE: "mllib_2.3",
            self.client.repository.ModelMetaNames.RUNTIME_UID: "spark-mllib_2.3",
            self.client.repository.ModelMetaNames.SPACE_UID: TestWMLClientWithSpark.space_uid
                       }

        published_model = self.client.repository.store_model(model=model_data['model'], meta_props=model_props, training_data=model_data['training_data'], pipeline=model_data['pipeline'])
        TestWMLClientWithSpark.model_uid = self.client.repository.get_model_uid(published_model)
        TestWMLClientWithSpark.logger.info("Published model ID:" + str(TestWMLClientWithSpark.model_uid))
        self.assertIsNotNone(TestWMLClientWithSpark.model_uid)

    def test_04_get_details(self):
        TestWMLClientWithSpark.logger.info("Get details")
        details = self.client.repository.get_details(self.model_uid)
        print(details)
        TestWMLClientWithSpark.logger.debug("Model details: " + str(details))
        self.assertTrue("Spark" in str(details))

    def test_09_delete_model(self):
        TestWMLClientWithSpark.logger.info("Delete model")
        self.client.repository.delete(TestWMLClientWithSpark.model_uid)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest,time

import logging
from ibm_watson_machine_learning.tests.Cloud.preparation_and_cleaning import *

# This is only for dev testing, since otherwise we need to everytime create cos buckets, create projects
# via api, etc. Should be done when possible though
class TestRshinyApp(unittest.TestCase):
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.project_id = get_project_id()

        self.client.set.default_project(self.project_id)

        print(self.client.service_instance.get_details())

        print("project_id: ", self.project_id)

    def test_01_create_shiny_asset(self):

        self.client.shiny.ConfigurationMetaNames.show()

        meta_prop_shiny = {
            self.client.shiny.ConfigurationMetaNames.NAME: "my shiny app project",
            self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "shiny app for project"
        }

        shiny_details = self.client.shiny.store(meta_prop_shiny, file_path="artifacts/app.R.zip")

        TestRshinyApp.shiny_asset_id = self.client.shiny.get_id(shiny_details)
        TestRshinyApp.shiny_asset_url = self.client.shiny.get_href(shiny_details)
        print("shiny asset ID:" + str(TestRshinyApp.shiny_asset_id))
        print("shiny asset URL:" + str(TestRshinyApp.shiny_asset_url))
        self.assertIsNotNone(TestRshinyApp.shiny_asset_id)
        self.assertIsNotNone(TestRshinyApp.shiny_asset_url)

    def test_02_download_shiny_content(self):
        try:
            os.remove('test_shiny_asset.zip')
        except:
            pass
        self.client.shiny.download(TestRshinyApp.shiny_asset_id, filename='test_shiny_asset.zip')
        try:
            os.remove('test_shiny_asset.zip')
        except:
            pass

    def test_03_get_details(self):

        details = self.client.shiny.get_details(TestRshinyApp.shiny_asset_id)
        print(details)
        self.assertTrue(TestRshinyApp.shiny_asset_id in str(details))

    def test_04_list(self):
        self.client.shiny.list()

    def test_05_revisions(self):
        meta_prop_shiny_revision1 = {
            self.client.shiny.ConfigurationMetaNames.NAME: "Shiny revision 1",
            self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "Shiny revision 1"
        }

        # Update meta and attachment for revision 1 creation
        shiny_update = self.client.shiny.update(TestRshinyApp.shiny_asset_id,
                                                meta_prop_shiny_revision1,
                                                file_path="artifacts/app1.R.zip")
        attachment_id = shiny_update[u'metadata'][u'attachment_id']

        # Create revision
        revision = self.client.shiny.create_revision(TestRshinyApp.shiny_asset_id)

        self.assertTrue(revision[u'metadata'][u'revision_id'] and revision[u'metadata'][u'revision_id'] == 1)

        new_attachment_id = revision[u'metadata'][u'attachment_id']

        self.assertTrue(attachment_id != new_attachment_id)

        # List revisions
        self.client.shiny.list_revisions(TestRshinyApp.shiny_asset_id)

        self.client.shiny.get_revision_details(TestRshinyApp.shiny_asset_id, 1)

        try:
            os.remove('test_shiny_asset_rev1.zip')
        except:
            pass

        # Download revision 1 attachment
        self.client.shiny.download(TestRshinyApp.shiny_asset_id, filename='test_shiny_asset_rev1.zip', rev_uid=1)
        try:
            os.remove('test_shiny_asset_rev1.zip')
        except:
            pass

    def test_10_delete_shiny_asset(self):
        TestRshinyApp.logger.info("Delete function")
        self.client.shiny.delete(TestRshinyApp.shiny_asset_id)

if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

class MetaProps:
    """
    Holder for props used during creation of ML artifacts.

    :param map meta: Map of pair key and value where key is taken from MetaNames.
    """
    def __init__(self, meta):
        self.meta = meta

    def available_props(self):
        """Return list of strings with names of available props."""
        return self.meta.keys()

    def prop(self, name):
        """Get prop value by name."""
        return self.meta.get(name)

    def merge(self, other):
        """Merge other MetaProp object to first one. Modify first MetaProp object, doesn't return anything."""
        self.meta.update(other.meta)

    def add(self, name, value):
        """
        Add new prop.

        :param str name: Key for value. Should be one of the values from MetaNames.
        :param object value: Any type of object
        """
        self.meta[name] = value

    def get(self):
        """returns meta prop dict"""
        return self.meta



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import ibm_boto3
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError
from ibm_watson_machine_learning.tests.utils import create_connection_to_cos
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestTSAsync, \
    AbstractTestWebservice

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics


class TestAutoAIRemote(AbstractTestTSAsync, unittest.TestCase):
    """
    The test can be run on Cloud only
    """

    cos_resource = None
    data_location = './autoai/data/Twitter_volume_AMZN_multivariate.csv'

    data_cos_path = 'Twitter_volume_AMZN_multivariate.csv'

    batch_payload_location = data_location
    batch_payload_cos_location = data_cos_path

    SPACE_ONLY = False
    BATCH_DEPLOYMENT = False

    OPTIMIZER_NAME = "Twitter_volume_AMZN_multivariate test sdk"
    DEPLOYMENT_NAME = OPTIMIZER_NAME + "Deployment"

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        prediction_type=PredictionType.FORECASTING,
        prediction_columns=['value1', 'value2'],
        timestamp_column_name='day',
        backtest_num=3,
        lookback_window=10,
        forecast_window=5,
        max_number_of_estimators=2,
        notebooks=True
    )

    def test_00b_prepare_COS_instance_and_connection(self):
        TestAutoAIRemote.connection_id, TestAutoAIRemote.bucket_name = create_connection_to_cos(
            wml_client=self.wml_client,
            cos_credentials=self.cos_credentials,
            cos_endpoint=self.cos_endpoint,
            bucket_name=self.bucket_name,
            save_data=True,
            data_path=self.data_location,
            data_cos_path=self.data_cos_path)

        self.assertIsInstance(self.connection_id, str)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(
            connection_asset_id=self.connection_id,
            location=S3Location(
                bucket=self.bucket_name,
                path=self.data_cos_path
            )
        )
        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNone(obj=TestAutoAIRemote.results_connection)

    def test_99_delete_connection(self):
        if not self.SPACE_ONLY:
            self.wml_client.set.default_project(self.project_id)

        self.wml_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.connections.get_details(self.connection_id)


if __name__ == '__main__':
    unittest.main()



__all__ = [
    'save_metadata_for_ui',
    'save_computed_pipelines_for_ui',
    'save_experiment_for_ui'
]

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os
from datetime import datetime as dt
from typing import Dict, List, TYPE_CHECKING
from uuid import uuid4

import pandas as pd
import requests

from .watson_studio import get_project

if TYPE_CHECKING:
    from ibm_watson_machine_learning import APIClient


def save_metadata_for_ui(wml_pipeline_config: Dict, training_details: Dict) -> None:
    """Prepare AutoAI experiment metadata file for UI.

    Parameters
    ----------
    wml_pipeline_config: dictionary, required
        Dictionary with WML pipeline configuration.

    training_details: dictionary, required
        Dictionary with training details.
    """
    experiment_id = training_details['entity']['results_reference']['location']['path'].split('/')[-2].split(
        'auto_ml.')[-1]
    parameters = wml_pipeline_config['entity']['document']['pipelines'][0]['nodes'][0]['parameters']
    learning_type = parameters['optimization']['learning_type']
    distinct_values = training_details['entity']['status'].get(
        'metrics', [{'context': {'classes': []}}])[0]['context']['classes']

    # note: mapping classification types to UI types
    if learning_type == 'classification':
        if len(distinct_values) == 2:
            learning_type = 'binary'
        else:
            learning_type = 'multiclass'
    # --- end note

    project = get_project()
    project_id = project.get_metadata()["metadata"]["guid"]

    training_data_references = training_details['entity']['training_data_references']

    creation_time = training_details['metadata']['created_at']
    timestamp = int(f"{int(dt.timestamp(dt.strptime(creation_time[:-5].replace('T', ' '), '%Y-%m-%d %H:%M:%S')))}000")

    # note: read associated WMLS data with project
    with open(os.path.join(os.path.abspath('.'), project_id, 'project.json'), 'r') as f:
        data = json.load(f)
    # --- end note

    service_instance_id = data['compute'][0]['guid']

    template = {
        "metadata": {
            "name": f"{wml_pipeline_config['metadata']['name']} SDK",
            "asset_type": "auto_ml",
            "origin_country": "us",
            "created_at": creation_time,
            "created": timestamp,
            "project_id": project_id,
            "asset_id": f"auto_ml.{experiment_id}",
            "sandbox_id": project_id,
            "last_modified": creation_time,
            "asset_attributes": [
                "auto_ml"
            ],
            "usage": {
                "last_update_time": timestamp,
                "last_updated_at": creation_time
            }
        },
        "entity": {
            "auto_ml": {
                "wml": {
                    "service_instance_id": service_instance_id,
                    "pipeline": {
                        "metadata": {
                            "name": wml_pipeline_config['metadata']['name'],
                            "guid": wml_pipeline_config['metadata']['guid'],
                            "tags": [
                                f"dsx-project.{project_id}"
                            ],
                            "id": wml_pipeline_config['metadata']['id'],
                            "modified_at": wml_pipeline_config['metadata']['modified_at'],
                            "created_at": wml_pipeline_config['metadata']['created_at'],
                            "owner": wml_pipeline_config['metadata']['owner'],
                            "href": wml_pipeline_config['metadata']['href'],
                            "space_id": wml_pipeline_config['metadata']['space_id']
                        },
                        "entity": {
                            "tags": [
                                {
                                    "value": f"dsx-project.{project_id}",
                                    "description": "guid of associated DSX project"
                                }
                            ],
                            "space": wml_pipeline_config['entity']['space'],
                            "name": wml_pipeline_config['entity']['name']
                        }
                    },
                    "training_definition": {
                        "metadata": {
                            "created_at": training_details['metadata']['created_at'],
                            "guid": training_details['metadata']['guid'],
                            "href": training_details['metadata']['href'],
                            "id": training_details['metadata']['id'],
                            "space_id": training_details['entity']['space_id'],
                            "tags": [
                                f"dsx-project.{project_id}"
                            ]
                        },
                        "entity": {
                            "pipeline": training_details['entity']['pipeline'],
                            "results_reference": training_details['entity']['results_reference'],
                            "space": training_details['entity']['space'],
                            "space_id": training_details['entity']['space_id'],
                            "tags": [
                                {
                                    "description": "guid of associated DSX project",
                                    "value": f"dsx-project.{project_id}"
                                }
                            ],
                            "training_data_references": training_data_references
                        }
                    }
                },
                "status": "completed",
                "learning_type": learning_type
            }
        },
        "attachments": [
            {
                "asset_type": "auto_ml",
                "name": "config",
                "description": f"config attachment for auto_ml asset auto_ml.{experiment_id}",
                "object_key": f"auto_ml.{experiment_id}/attachments/config_auto_ml.{experiment_id}",
                "object_key_is_read_only": True,
                "created_at": creation_time,
                "id": f"{uuid4()}",
                "is_managed": False,
                "is_referenced": True,
                "is_remote": False
            },
            {
                "asset_type": "auto_ml",
                "name": "referenced",
                "description": f"referenced attachment for auto_ml asset auto_ml.{experiment_id}",
                "object_key": f"auto_ml.{experiment_id}/attachments/referenced_auto_ml.{experiment_id}",
                "object_key_is_read_only": True,
                "created_at": creation_time,
                "id": f"{uuid4()}",
                "is_managed": False,
                "is_referenced": True,
                "is_remote": False
            }
        ]
    }

    file_name = os.path.join(os.path.abspath('.'), project_id, 'assets', '.METADATA', f"auto_ml.{experiment_id}.json")

    with open(file_name, 'w') as f:
        f.write(json.dumps(template))


def save_computed_pipelines_for_ui(wml_client: 'APIClient', run_id: str) -> None:
    """Download all AutoAI experiments files into the local storage for UI usage.

    Parameters
    ----------
    wml_client: APIClient, required

    run_id: str, required
        ID of the training.
    """
    print('Saving all computed pipelines locally for WSD...')

    models_outputs = (
        'cognito_output',
        'hpo_c_output',
        'hpo_d_output',
        'pre_hpo_d_output'
    )

    project = get_project()
    project_id = project.get_metadata()["metadata"]["guid"]

    training_details = wml_client.training.get_details(run_id)

    # note: prepare needed information to fetch data
    path = training_details['entity']['results_reference']['location']['path'].split('assets/auto_ml/')[-1]
    experiment_id = path.split('/')[0].split('auto_ml.')[-1]
    remote_path = f"{path}/{run_id}/data/automl/"
    # --- end note

    # note: prepare local path within WSD
    local_path = os.path.join(os.path.abspath('.'), project_id, 'assets', 'auto_ml', f"auto_ml.{experiment_id}",
                              'wml_data', run_id, 'data', 'automl')
    # --- end note

    # note: download each model output (models, jsons etc.)
    for output in models_outputs:
        generated_path = f"{remote_path}{output}"

        response = requests.get(
            url=f"{wml_client.service_instance._href_definitions.get_wsd_model_attachment_href()}auto_ml/{generated_path}",
            headers=wml_client._get_headers(),
            params={'space_id': wml_client.default_space_id,
                    'flat': 'true'},
            verify=False)

        response_with_paths = wml_client.data_assets._handle_response(200, u'getting data from WMLS', response)

        for resource in response_with_paths['resources']:
            pipeline_dir = resource['path'].split('/')[-2]
            file_name = resource['path'].split('/')[-1]

            generated_local_path = os.path.join(local_path, output, pipeline_dir)
            create_dir(generated_local_path)

            response_with_file = requests.get(
                url=f"{wml_client.service_instance._href_definitions.get_wsd_model_attachment_href()}auto_ml/{generated_path}/{pipeline_dir}/{file_name}",
                headers=wml_client._get_headers(),
                params={'space_id': wml_client.default_space_id},
                verify=False)

            with open(os.path.join(generated_local_path, file_name), 'wb') as f:
                f.write(response_with_file.content)
    # --- end note
    print("Saved.")


def save_experiment_for_ui(wml_pipeline_config: Dict,
                           training_details: Dict,
                           training_data_file_path: str,
                           asset_id: str,
                           asset_name: str) -> None:
    """Prepare AutoAI experiment configuration from wml pipeline config for WSD UI reuse.

    Parameters
    ----------
    wml_pipeline_config: dictionary, required
        Dictionary with WML pipeline configuration.

    training_details: dictionary, required
        Dictionary with training details.

    training_data_file_path: str, required
        Local path to the training data asset.

    asset_id: str, required
        ID of the data asset on WSD.

    asset_name: str, required
        Name of the data asset on WSD.

    Returns
    -------
    Dictionary with AutoAI experiment configuration for WSD UI.
    """
    print('Saving experiment metadata locally for WSD...')

    experiment_id = training_details['entity']['results_reference']['location']['path'].split('/')[-2].split(
        'auto_ml.')[-1]

    config = prepare_config_auto_ml(wml_pipeline_config, training_details, training_data_file_path,
                                    asset_id, asset_name)
    reference = prepare_reference_auto_ml(wml_pipeline_config, training_details)

    project = get_project()
    project_id = project.get_metadata()["metadata"]["guid"]

    config_path = os.path.join(os.path.abspath('.'), project_id, 'assets',
                               'auto_ml', f"auto_ml.{experiment_id}", 'attachments')
    config_file_name = f"config_auto_ml.{experiment_id}"
    reference_path = os.path.join(os.path.abspath('.'), project_id, 'assets',
                                  'auto_ml', f"auto_ml.{experiment_id}", 'attachments')
    reference_file_name = f"referenced_auto_ml.{experiment_id}"

    # note: make sure that directories are created
    create_dir(config_path)
    create_dir(reference_path)
    # --- end note

    with open(os.path.join(config_path, config_file_name), 'w') as f:
        f.write(decode_for_ui(config))

    with open(os.path.join(reference_path, reference_file_name), 'w') as f:
        f.write(decode_for_ui(reference))

    print("Saved.")


def create_dir(path: str) -> None:
    """Creates entire path of directories if missing."""
    try:
        os.mkdir(path)

    except FileExistsError:
        pass

    except FileNotFoundError:
        create_dir(f'{os.sep}'.join(path.split(os.sep)[:-1]))
        create_dir(path)


def decode_for_ui(data: Dict) -> str:
    """Convert / decode dictionary with data for UI accessible format."""
    _dict = {
        "name": data['name'],
        "body": 'placeholder'
    }
    _dict = json.dumps(_dict)
    _body = json.dumps(data['body']).replace(r'"', r'\"')
    _dict = _dict.replace('placeholder', _body)

    return _dict


def prepare_config_auto_ml(wml_pipeline_config: Dict,
                           training_details: Dict,
                           training_data_file_path: str,
                           asset_id: str,
                           asset_name: str) -> Dict:
    """Prepare AutoAI experiment configuration from wml pipeline config for WSD UI reuse.

    Parameters
    ----------
    wml_pipeline_config: dictionary, required
        Dictionary with WML pipeline configuration.

    training_details: dictionary, required
        Dictionary with training details.

    training_data_file_path: str, required
        Local path to the training data asset.

    asset_id: str, required
        ID of the data asset on WSD.

    asset_name: str, required
        Name of the data asset on WSD.

    Returns
    -------
    Dictionary with AutoAI experiment configuration for WSD UI.
    """
    experiment_id = training_details['entity']['results_reference']['location']['path'].split('/')[-2].split(
        'auto_ml.')[-1]
    parameters = wml_pipeline_config['entity']['document']['pipelines'][0]['nodes'][0]['parameters']

    try:
        spec = wml_pipeline_config['entity']['document']['runtimes'][0]['app_data']['wml_data']['hardware_spec']['name']

    except KeyError:
        spec = wml_pipeline_config['entity']['document']['runtimes'][0][
            'app_data']['wml_data']['runtime_spec_v4']['compute']['name']

    name = os.path.join('auto_ml', f"auto_ml.{experiment_id}", 'attachments', f"config_auto_ml.{experiment_id}")
    label_col = parameters['optimization']['label']
    learning_type = parameters['optimization']['learning_type']
    scorer_for_ranking = parameters['optimization']['scorer_for_ranking']
    compute_plan = spec.lower()
    holdout_param = parameters['optimization']['holdout_param']
    daub_include_only_estimators = parameters['optimization']['daub_include_only_estimators']
    cognito_transform_names = parameters['optimization']['cognito_transform_names']
    max_num_daub_ensembles = int(parameters['optimization']['max_num_daub_ensembles'])
    distinct_values = training_details['entity']['status'].get(
        'metrics', [{'context': {'classes': []}}])[0]['context']['classes']

    # note: mapping classification types to UI types
    if learning_type == 'classification':
        if len(distinct_values) == 2:
            learning_type = 'binary'
        else:
            learning_type = 'multiclass'
    # --- end note

    # note: asset information part
    data = pd.read_csv(training_data_file_path, nrows=2)
    label_col_idx = data.columns.to_list().index(label_col)
    total_columns = len(data.columns)
    asset_size = os.stat(training_data_file_path).st_size
    # --- end note

    template = {
        "name": name,
        "body": {
            "target": {
                "label_col": label_col,
                "label_col_idx": label_col_idx,
                "learning_type": learning_type,
                "metric": scorer_for_ranking,
                "main_table_id": asset_id,
                "distinct_values": distinct_values
            },
            "source_assets": [
                {
                    "asset_id": asset_id,
                    "asset_type": "data_asset",
                    "asset_name": asset_name,
                    "asset_size": asset_size,
                    "total_columns": total_columns
                }
            ],
            "obm": {},
            "kb": {
                "run_settings": {
                    "compute_plan": compute_plan,
                    "holdout_param": holdout_param,
                    "daub_include_only_estimators": daub_include_only_estimators,
                    "cognito_transform_names": cognito_transform_names,
                    "max_num_daub_ensembles": max_num_daub_ensembles
                }
            },
            "column_settings": {}
        }
    }

    return template


def prepare_reference_auto_ml(wml_pipeline_config: Dict, training_details: Dict) -> Dict:
    """Prepare AutoAI experiment configuration from wml pipeline config for WSD UI reuse.

    Parameters
    ----------
    wml_pipeline_config: dictionary, required
        Dictionary with WML pipeline configuration.

    training_details: dictionary, required
        Dictionary with training details.

    Returns
    -------
    Dictionary with AutoAI experiment configuration for WSD UI.
    """

    completed_at = training_details['entity']['status']['completed_at']
    last_modified = training_details['metadata']['modified_at']
    created_at = training_details['metadata']['created_at']
    guid = training_details['metadata']['guid']
    experiment_id = training_details['entity']['results_reference']['location']['path'].split('/')[-2].split(
        'auto_ml.')[-1]
    name = os.path.join('auto_ml', f"auto_ml.{experiment_id}", 'attachments', f"referenced_auto_ml.{experiment_id}")
    message_log = extract_message_log(training_details)
    estimators = extract_estimators(wml_pipeline_config, training_details)

    template = {
        "name": name,
        "body": {
            "version": 2,
            "trainingGuid": guid,
            "message": message_log[-1],
            "state": "completed",
            "pipelines": extract_pipelines(training_details),
            "estimators": estimators,
            "created_at": created_at,
            "preprocessing_step": "AUTO_AI_STEP_COMPLETED",
            "message_log": message_log,
            "completed_at": completed_at,
            "last_modified": last_modified
        }
    }

    return template


def extract_pipelines(training_details: Dict) -> Dict:
    """Extract information about computed pipelines from training details.
    Parameters
    ----------
    training_details: dictionary, required
        Dictionary with training details.

    Returns
    -------
    Dictionary with AutoAI pipelines details.
    """
    pipelines = {}
    stages = training_details['entity']['status'].get('metrics', [])

    current_pipeline_name = None
    for stage in stages[::-1]:
        name = stage['context']['intermediate_model']['name']

        if current_pipeline_name != name:
            current_pipeline_name = name

            pipelines[name] = stage

    sorted_pipelines = {}
    for key, value in sorted(pipelines.items()):
        sorted_pipelines[key] = value

    return sorted_pipelines


def extract_estimators(wml_pipeline_config: Dict, training_details: Dict) -> Dict:
    """Extract information about estimators from training details.
    Parameters
    ----------
    wml_pipeline_config: dictionary, required
        Dictionary with WML pipeline configuration.

    training_details: dictionary, required
        Dictionary with training details.

    Returns
    -------
    Dictionary with estimators and scores.
    """
    estimators = {}
    parameters = wml_pipeline_config['entity']['document']['pipelines'][0]['nodes'][0]['parameters']
    scorer_for_ranking = parameters['optimization']['scorer_for_ranking']
    stages = training_details['entity']['status'].get('metrics', [])

    current_pipeline_estimator = None
    for stage in stages[::-1]:
        estimator_name = stage['context']['intermediate_model']['pipeline_nodes'][-1]
        score = stage['ml_metrics'][f'training_{scorer_for_ranking}']

        if current_pipeline_estimator != estimator_name:
            current_pipeline_estimator = estimator_name

            estimators[estimator_name] = {'score': score}

    return estimators


def extract_message_log(training_details: Dict) -> List[Dict]:
    """Creates message log for training.
    Parameters
    ----------
    training_details: dictionary, required
        Dictionary with training details.

    Returns
    -------
    List of dictionaries with messages.
    """
    steps = {'DAUB': 'feature_engineering',
             'hpo_d': 'model_optimization',
             'cognito': 'transforming_data',
             'hpo_c': 'optimizing_hyperparams'}

    message_log = []
    stages = training_details['entity']['status'].get('metrics', [])

    timestamp = 0
    model_name = 'P4'
    duration = 1

    for stage in stages:
        model_name = stage['context']['intermediate_model']['name']
        timestamp = int(f"{int(dt.timestamp(dt.strptime(training_details['metadata']['created_at'][:-5].replace('T', ' '), '%Y-%m-%d %H:%M:%S')))}000")
        step = steps[stage['context']['intermediate_model']['composition_steps'][-1]]
        duration = stage['context']['intermediate_model']['duration']

        message_start = {
            "timestamp": timestamp,
            "text": f"Train_log_msg_{step}_start",
            "running_pipeline": model_name,
            "level": "info"
        }

        message_end = {
            "timestamp": timestamp + duration,
            "text": f"Train_log_msg_{step}_complete",
            "running_pipeline": model_name,
            "level": "info"
        }

        message_log.append(message_start)
        message_log.append(message_end)

    message_log.append(
        {
            "timestamp": timestamp + duration,
            "text": "Train_log_msg_data_training_complete",
            "running_pipeline": model_name,
            "level": "info"
        }
    )
    message_log.append(
        {
            "timestamp": timestamp + duration,
            "text": "autoai execution completed",
            "running_pipeline": model_name,
            "level": "info"
        }
    )
    message_log.append(
        {
            "timestamp": timestamp + duration,
            "text": "Train_log_msg_data_training_complete",
            "running_pipeline": model_name,
            "level": "info"
        }
    )
    message_log.append(
        {
            "timestamp": timestamp + duration,
            "text": "Train_log_msg_pipeline_created",
            "running_pipeline": model_name,
            "level": "info"
        }
    )
    message_log.append(
        {
            "timestamp": timestamp + duration,
            "text": "Train_log_msg_training_completed",
            "running_pipeline": model_name,
            "level": "info"
        }
    )

    return message_log



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import ibm_boto3
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket, is_cp4d, save_data_to_cos_bucket
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestAutoAISync,\
    AbstractTestWebservice

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics


class TestAutoAIRemote(AbstractTestAutoAISync, AbstractTestWebservice, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/xlsx/CarPrice_bank__two_sheets.xlsx'
    sheet_name = 'bank'
    sheet_number = 1

    cos_endpoint = "https://s3.us-south.cloud-object-storage.appdomain.cloud"
    data_cos_path = 'data/CarPrice_bank__two_sheets.xlsx'

    SPACE_ONLY = True

    OPTIMIZER_NAME = "Bank test sdk"

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='test description',
        prediction_type=PredictionType.BINARY,
        prediction_column='y',
        positive_label='yes',
        scoring=Metrics.ACCURACY_SCORE,
        holdout_size=0.1,
        max_number_of_estimators=1,
        excel_sheet=sheet_name
    )

    def test_00b_prepare_COS_instance(self):
        TestAutoAIRemote.bucket_name = save_data_to_cos_bucket(self.data_location,
                                                               self.data_cos_path,
                                                               access_key_id=self.cos_credentials['cos_hmac_keys'][
                                                                   'access_key_id'],
                                                               secret_access_key=self.cos_credentials['cos_hmac_keys'][
                                                                   'secret_access_key'],
                                                               cos_endpoint=self.cos_endpoint,
                                                               bucket_name=self.bucket_name)

    def test_00c_prepare_connection_to_COS(self):
        connection_details = self.wml_client.connections.create({
            'datasource_type': self.wml_client.connections.get_datasource_type_uid_by_name('bluemixcloudobjectstorage'),
            'name': 'Connection to COS for tests',
            'properties': {
                'bucket': self.bucket_name,
                'access_key': self.cos_credentials['cos_hmac_keys']['access_key_id'],
                'secret_key': self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                'iam_url': self.wml_client.service_instance._href_definitions.get_iam_token_url(),
                'url': self.cos_endpoint
            }
        })

        TestAutoAIRemote.connection_id = self.wml_client.connections.get_uid(connection_details)
        self.assertIsInstance(self.connection_id, str)

    def test_00d_prepare_connected_data_asset(self):
        asset_details = self.wml_client.data_assets.store({
            self.wml_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: self.connection_id,
            self.wml_client.data_assets.ConfigurationMetaNames.NAME: "Bank - training asset",
            self.wml_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: self.data_cos_path
        })

        TestAutoAIRemote.asset_id = self.wml_client.data_assets.get_id(asset_details)
        self.assertIsInstance(self.asset_id, str)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(data_asset_id=self.asset_id)
        TestAutoAIRemote.results_connection = DataConnection(
            connection_asset_id=self.connection_id,
            location=S3Location(
                bucket=self.bucket_name,
                path=self.results_cos_path
            )
        )

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNotNone(obj=TestAutoAIRemote.results_connection)

    def test_99_delete_connection_and_connected_data_asset(self):
        self.wml_client.data_assets.delete(self.asset_id)
        self.wml_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.data_assets.get_details(self.asset_id)
            self.wml_client.connections.get_details(self.connection_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging

from  ibm_watson_machine_learning.libs.repo.mlrepository.artifact_reader import ArtifactReader

logger = logging.getLogger('RuntimesArtifactReader')


class RuntimesArtifactReader(ArtifactReader):
    def __init__(self, runtimespec_path):
        self.runtimespec_path = runtimespec_path

    def read(self):
        return self._open_stream()

    # This is a no. op. for RuntimeYmlFileReader as we do not want to delete the
    # archive file.
    def close(self):
        pass

    def _open_stream(self):
        return open(self.runtimespec_path, 'rt')




#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc
from pyspark import SparkContext
from pyspark.sql import DataFrame,SQLContext

from IBMSparkPipelineModel import IBMSparkPipelineModel
from Wrapper import PythonJavaConversions

class Result(object):
    """
    The absract base class for Result. The `stage` field is a string representing
    the end stage of the DAG.
    """
    __metaclass__ = abc.ABCMeta

    stage = None

    @abc.abstractmethod
    def _populateResult(self, jlist):
        pass


class DataResult(Result):
    """
    The DataResult class extends from :class:`Result`. The `data` field is a DataFrame.
    """
    data = None
    _sc = SparkContext._active_spark_context

    @classmethod
    def _populateResult(cls, jlist):
        resultlist = [DataResult() for index in range(len(jlist))]
        for index in range(len(jlist)):
            resultlist[index].stage = str(jlist[index].stage())
            resultlist[index].data = DataFrame(jlist[index].data(), SQLContext.getOrCreate(cls._sc))
        return resultlist

    def show(self):
        """
        Prints the DataFrame field of DataResult class.

        :return: None
        """
        self.data.show()


class ModelResult(Result):
    """
    The ModelResult class extends from :class:`Result`. The `pipelineModel` field is an object of type
    :class:`pipeline.IBMSparkPipelineModel.IBMSparkPipelineModel`.
    """
    pipelineModel = None

    @classmethod
    def _populateResult(cls, jlist):
        resultlist = [ModelResult() for index in range(len(jlist))]
        for index in range(len(jlist)):
            resultlist[index].stage = str(jlist[index].stage())
            resultlist[index].pipelineModel = cls._from_java(jlist[index].pipelineModel())
        return resultlist

    @classmethod
    def _from_java(cls, java_stage):
        py_stages = [PythonJavaConversions._from_java_stage(s) for s in java_stage.stages()]
        # Create a new instance of this stage.
        py_stage = IBMSparkPipelineModel(py_stages)
        py_stage = PythonJavaConversions._resetUid(py_stage, java_stage.uid())
        return py_stage


class SinkResult(Result):
    """
    The SinkResult class extends from :class:`Result`. In case of a SinkResult, only the `stage`
    field is returned.
    """
    @classmethod
    def _populateResult(self, jlist):
        resultlist = [SinkResult() for index in range(len(jlist))]
        for index in range(len(jlist)):
            resultlist[index].stage = str(jlist[index].stage())
        return resultlist



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import certifi
import json
from ibm_watson_machine_learning.libs.repo.swagger_client.rest import ApiException

try:
    import urllib3
except ImportError:
    raise ImportError('urllib3 is missing')

try:
    # for python3
    from urllib.parse import urlencode
except ImportError:
    # for python2
    from urllib import urlencode

http = urllib3.PoolManager(
    cert_reqs='CERT_REQUIRED',
    ca_certs=certifi.where()
)


class MLAuthorization(object):
    """
    Class used to authorize user.

    MLAuthorization shouldn't be used alone. It is created to be used as a base class by MLRepositoryClient.
    """
    def __init__(self, api_client):
        self.api_client = api_client
        self.stage_url = "https://iam.stage1.bluemix.net/oidc/token"
        self.prod_url = "https://iam.bluemix.net/oidc/token"


    def _add_header(self, key, value):
        self.api_client.set_default_header(key, value)
        self.api_client.set_default_header(key, value)


    def iam_authorize(self,vcap):
        url         =   vcap.get('url')
        api_key     =   vcap.get('api_key')
        instance_id =   vcap.get('instance_id')
        user        =   vcap.get('username')
        password    =   vcap.get('password')

        if api_key is None:
            if user is None:
                raise TypeError("Watson ML service credentials: username not defined ")
            elif password is None:
                raise TypeError("Watson ML service credentials: password  not defined")
            else:
                #Authorize with ML Token
                self.authorize(user,password)
        elif instance_id is None:
            raise TypeError("Watson ML service credentials: instance_id not defined")
        else:
            #Authorize IAM Token
            self.generate_iam_token(url,api_key,instance_id)


    def get_iam_url(self, base_service_url) :
        path = base_service_url

        paramList = path.split('.')
        if paramList[1] == "stage1":
            iam_url= self.stage_url
        else:
            iam_url= self.prod_url
        return iam_url

    def generate_iam_token(self,base_url, api_key,instance_id):
        oidc_data = {'grant_type': 'urn:ibm:params:oauth:grant-type:apikey',
                     'response_type': 'cloud_iam',
                     'apikey': api_key}
        headers = {'Content-Type':'application/x-www-form-urlencoded'}
        url = self.get_iam_url(base_url)

        try:
            response = http.request(
                'POST',
                url,
                fields=oidc_data,
                encode_multipart=False,
                headers=headers
            )
            if response.status == 200:
                iam_token = json.loads(response.data.decode('UTF-8')).get('access_token')
                self.authorize_with_iamtoken(iam_token,instance_id)
            elif response.status == 401:
                raise ApiException(
                    status=401,
                    reason="The incoming request did not contain a valid authentication information"
                )
            elif response.status == 403:
                raise ApiException(
                    status=403,
                    reason="The incoming request did not contain a valid authentication information."
                )
            elif response.status == 400:
                raise ApiException(
                    status=400,
                    reason="Parameter validation failed."
                )
            elif response.status == 500:
                raise ApiException(
                    status=500,
                    reason="Internal Server error."
                )
            else:
                msg = "Unable to authenticate due to error :\n" + "{0}\n{1}".format(response.status, response.text)
                raise ApiException(
                    status=404,
                    reason=msg
                )
        except ApiException as e:
            raise e

        except BaseException as e:
            msg = "Unable to authenticate due to error :\n" + "{0}\n{1}".format(type(e).__name__, str(e))
            raise ApiException(
                status=404,
                reason=msg
            )

    def authorize(self, user, password):
        """
        Authorizes user by username and password

        Both username and password should be credentials taken from Watson Machine Learning instance.

        :param str user: username taken from WML instance
        :param str password: password corresponding with user name taken from WML instance

        """
        headers = urllib3.util.make_headers(basic_auth='{}:{}'.format(user, password))
        url = '{}/v2/identity/token'.format(self.api_client.repository_path)
        try:
            response = http.request(
                      'GET',
                      url,
                      headers=headers
                     )
            if response.status == 200:
                token = json.loads(response.data.decode('UTF-8')).get('token')
                self.authorize_with_token(token)
            elif response.status == 401:
                raise ApiException(
                    status=404,
                    reason="Invalid credentials. Authorization failed."
                    )
            elif response.status == 403:
                raise ApiException(
                    status=404,
                    reason="Authentication failed due to insufficient permission"
                    )
            else:
                msg = "Unable to authenticate due to error :\n" + "{0}\n{1}".format(response.status, response.text)
                raise ApiException(
                    status=404,
                    reason=msg
                   )
        except ApiException as e:
            raise e

        except BaseException as e:
            msg = "Unable to authenticate due to error :\n" + "{0}\n{1}".format(type(e).__name__, str(e))
            raise ApiException(
                status=404,
                reason=msg
            )

    def authorize_with_token(self, token):
        """
            Authorizes user by token.

            Token should be generated during basic authentication using username and password from WML service instance.

            :param str token: token

            """
        self._add_header('Authorization', 'Bearer {}'.format(token))


    def authorize_with_iamtoken(self, iamtoken,instance_id, platform_spaces=False):

        self._add_header('Authorization', 'Bearer {}'.format(iamtoken))
        # Cloud Convergence
        if not platform_spaces:
           self._add_header('ML-Instance-ID', instance_id)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import time

import logging
from ibm_watson_machine_learning.tests.ICP.preparation_and_cleaning import *
from ibm_watson_machine_learning.tests.ICP.models_preparation import *


class TestPkgExtn(unittest.TestCase):
    runtime_uid = None
    deployment_uid = None
    function_uid = None
    scoring_url = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestPkgExtn.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.space = self.client.spaces.store({self.client.spaces.ConfigurationMetaNames.NAME: "test_case_TestPkgExtnSw"})
        self.space_id = self.client.spaces.get_uid(self.space)
        self.client.set.default_space(self.space_id)

    # def test_01_service_instance_details(self):
    #     TestPkgExtn.logger.info("Check client ...")
    #     self.assertTrue(self.client.__class__.__name__ == 'APIClient')
    #
    #     TestPkgExtn.logger.info("Getting instance details ...")
    #     details = self.client.service_instance.get_details()
    #     TestPkgExtn.logger.debug(details)
    #
    #     self.assertTrue("published_models" in str(details))
    #     self.assertEqual(type(details), dict)


    def test_00_create_pkg_extn(self):


        meta_prop_pkg_extn = {
            self.client.package_extensions.ConfigurationMetaNames.NAME: "Pkg extension for conda",
            self.client.package_extensions.ConfigurationMetaNames.DESCRIPTION: "Pkg extension for conda",
            self.client.package_extensions.ConfigurationMetaNames.TYPE: "conda_yml",
        }

        pkg_extn_details = self.client.package_extensions.store(meta_props=meta_prop_pkg_extn,
                                                                file_path="artifacts/conda_regex.yaml")

        TestPkgExtn.pkg_extn_uid = self.client.package_extensions.get_uid(pkg_extn_details)
        pkg_extn_url = self.client.package_extensions.get_href(pkg_extn_details)
        TestPkgExtn.logger.info("sw spec ID:" + str(TestPkgExtn.pkg_extn_uid))
        TestPkgExtn.logger.info("w spec URL:" + str(pkg_extn_url))
        self.assertIsNotNone(TestPkgExtn.pkg_extn_uid)
        self.assertIsNotNone(pkg_extn_url)

    def test_01_get_details(self):

        details = self.client.package_extensions.get_details(TestPkgExtn.pkg_extn_uid)
        self.assertTrue("Pkg extension for conda" in str(details))


    def test_02_list_package_extensions(self):
        self.client.package_extensions.list()


    def test_03_create_sw_spec(self):

        self.client.software_specifications.ConfigurationMetaNames.show()
        base_sw_spec_uid = self.client.software_specifications.get_uid_by_name("ai-function_0.1-py3.6")

        meta_prop_sw_spec = {
            self.client.software_specifications.ConfigurationMetaNames.NAME: "sk_learn19_vshasha_new " + time.asctime(),
            self.client.software_specifications.ConfigurationMetaNames.DESCRIPTION: "Software specification for vshasha",
            self.client.software_specifications.ConfigurationMetaNames.BASE_SOFTWARE_SPECIFICATION: {"guid": base_sw_spec_uid}
        }

        sw_spec_details = self.client.software_specifications.store(meta_props=meta_prop_sw_spec)

        TestPkgExtn.sw_spec_uid = self.client.software_specifications.get_uid(sw_spec_details)
        sw_spec_url = self.client.software_specifications.get_href(sw_spec_details)
        TestPkgExtn.logger.info("sw spec ID:" + str(TestPkgExtn.sw_spec_uid))
        TestPkgExtn.logger.info("sw spec URL:" + str(sw_spec_url))
        self.assertIsNotNone(TestPkgExtn.sw_spec_uid)
        self.assertIsNotNone(sw_spec_url)


    def test_04_add_pkg_extn_to_sw_spec(self):

        self.client.software_specifications.add_package_extension(TestPkgExtn.sw_spec_uid, TestPkgExtn.pkg_extn_uid)


    def test_05_get_details(self):

        details = self.client.software_specifications.get_details(TestPkgExtn.sw_spec_uid)
        self.assertTrue("sk_learn19_vshasha_new" in str(details))
        self.assertTrue("Pkg extension for conda" in str(details))



    def test_06_del_pkg_extn_to_sw_spec(self):
        self.client.software_specifications.delete_package_extension(TestPkgExtn.sw_spec_uid, TestPkgExtn.pkg_extn_uid)


    def test_07_get_details(self):

        details = self.client.software_specifications.get_details(TestPkgExtn.sw_spec_uid)
        self.assertTrue("sk_learn19_vshasha_new" in str(details))
        self.assertTrue("Pkg extension for conda" not in str(details))


    def test_08_list(self):
        self.client.software_specifications.list()

    def test_09_delete_pkg_extn(self):
        TestPkgExtn.logger.info("Delete deployment")
        self.client.package_extensions.delete(TestPkgExtn.pkg_extn_uid)

    def test_10_delete_sw_spec(self):
        TestPkgExtn.logger.info("Delete deployment")
        self.client.software_specifications.delete(TestPkgExtn.sw_spec_uid)

    def test_11_delete_space(self):
        TestPkgExtn.logger.info("Delete space")
        self.client.spaces.delete(TestPkgExtn.space_id)



if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import uuid

from ibm_watson_machine_learning.helpers.connections import DataConnection, ContainerLocation
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket, is_cp4d, save_data_to_container
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestAutoAIAsync, \
    AbstractTestWebservice, AbstractTestBatch

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, ClassificationAlgorithms


class TestAutoAIRemoteFairness(AbstractTestAutoAIAsync, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/drug_train_data_updated.csv'

    data_cos_path = 'data/drug_train_data_updated.csv'

    SPACE_ONLY = False

    OPTIMIZER_NAME = "Drug data test sdk"

    HISTORICAL_RUNS_CHECK = False

    target_space_id = None

    fairness_info = {
        # "protected_attributes": [
        #     {"feature": "SEX", "privileged_groups": ['F']},
        #     {"feature": "BP", "privileged_groups": ["LOW", "NORMAL"]}
        # ],
        "favorable_labels": ["drugA", "drugC"],
        "protected_attributes": [],
    }

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='FAIRNESS experiment',
        prediction_type=PredictionType.MULTICLASS,
        prediction_column='DRUG',
        scoring="accuracy_and_disparate_impact",
        include_only_estimators=[ClassificationAlgorithms.LGBM,
                                 ClassificationAlgorithms.XGB],
        fairness_info=fairness_info,
        max_number_of_estimators=2,
        text_processing=False
    )

    def test_00b_write_data_to_container(self):
        if self.SPACE_ONLY:
            self.wml_client.set.default_space(self.space_id)
        else:
            self.wml_client.set.default_project(self.project_id)

        save_data_to_container(self.data_location, self.data_cos_path, self.wml_client)

    def test_00c_check_fairness_info(self):
        print(self.fairness_info)

        self.assertIn('favorable_labels', self.fairness_info)
        self.assertIsInstance(self.fairness_info['favorable_labels'], list)

    def test_02_data_reference_setup(self):
        TestAutoAIRemoteFairness.data_connection = DataConnection(
            location=ContainerLocation(path=self.data_cos_path
                                       ))
        TestAutoAIRemoteFairness.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemoteFairness.data_connection)
        self.assertIsNone(obj=TestAutoAIRemoteFairness.results_connection)

    def test_10_summary_listing_all_pipelines_from_wml(self):
        TestAutoAIRemoteFairness.summary = self.remote_auto_pipelines.summary()
        print(TestAutoAIRemoteFairness.summary)
        for feature in ['AGE', 'SEX']:
            self.assertIn(f'holdout_disparate_impact_{feature}', list(TestAutoAIRemoteFairness.summary.columns))
            self.assertIn(f'training_disparate_impact_{feature}', list(TestAutoAIRemoteFairness.summary.columns))

        for col in self.summary.columns:
            print(self.summary[col])

    def test_99_delete_connection_and_connected_data_asset(self):
        if not self.SPACE_ONLY:
            self.wml_client.set.default_project(self.project_id)
        self.wml_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.connections.get_details(self.connection_id)


class TestFairnessNumericalValues(TestAutoAIRemoteFairness):
    fairness_info = {
        "protected_attributes": [
            # {"feature": "AGE", "reference_group": [18, [38, 50], 60]},
            {"feature": "CHOLESTEROL", "reference_group": [[0, 0.6], [0.613261,0.613261]]}
        ],
        "favorable_labels": ["drugA"]
    }
    experiment_info = dict(
        name="Fairness",
        desc='FAIRNESS experiment',
        prediction_type=PredictionType.MULTICLASS,
        prediction_column='DRUG',
        scoring=Metrics.ACCURACY_SCORE,
        include_only_estimators=[ClassificationAlgorithms.SnapDT],
        fairness_info=fairness_info,
        notebooks=True,
        max_number_of_estimators=1,
        text_processing=False
    )



if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import importlib

from warnings import warn
import os
import sys

# Added fix for message "Using TensorFlow backend."

stderr = sys.stderr
try:
    sys.stderr = open(os.devnull, 'w')
    import keras
except Exception as ex:
    pass
sys.stderr = stderr


latest_required_libraries = ["tensorflow==2.4.3", "sklearn==0.23.2", "torch==1.7.1",
                             "keras==2.2.4", "numpy", "pandas", "pytest",
                             "yaml", "parse", "websockets", "jsonpickle", "requests",
                             "scipy==1.4.1", "environs", "pathlib2", "diffprivlib",
                             "psutil", "setproctitle", "tabulate", "lz4",
                             "cv2", "gym", "ray==0.8.0", "cloudpickle==1.3.0", "image"]

oldest_required_libraries = ["tensorflow==2.1.0", "sklearn==0.23.1", "torch==1.7.1",
                             "keras==2.2.4", "numpy==1.17.4", "pandas", "pytest",
                             "yaml", "parse", "websockets", "jsonpickle", "requests",
                             "scipy==1.4.1", "environs", "pathlib2", "diffprivlib",
                             "psutil", "setproctitle", "tabulate", "lz4",
                             "cv2", "gym", "ray==0.8.0", "cloudpickle==1.3.0", "image"]


for old_lib, new_lib in zip(oldest_required_libraries, latest_required_libraries):

    req_lib_new = new_lib.split("==")
    req_lib_old = old_lib.split("==")

    try:
        var = importlib.import_module(req_lib_new[0])
        if len(req_lib_old) == 2 and len(req_lib_new) == 2:
            if var.__version__ != req_lib_old[1] and var.__version__ != req_lib_new[1]:
                print("Module", req_lib_old[0], "has incompatible version:", var.__version__, "Oldest version is:", req_lib_old[1],
                      "but latest version is:", req_lib_new[1])

    except Exception as e:
        if len(req_lib_old) == 2 or len(req_lib_new) == 2:
            print("Missing module", "'" + req_lib_old[0] + "'", ", Oldest version is:", req_lib_old[1], "but latest version is: ", req_lib_new[1])
            print("With error message: " , e, "\n")
        else:
            print(e)
        continue



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import os

import ibm_boto3
from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location
from ibm_watson_machine_learning.tests.utils import create_connection_to_cos
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestOBM

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, ClassificationAlgorithms


class TestAutoAIRemote(AbstractTestOBM, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """
    HISTORICAL_RUNS_CHECK = False
    cos_resource = None

    subnames = ['insurance', 'customer', 'claim']

    deployment_payload_filenames = [f'{name}.csv' for name in subnames]
    input_data_filenames = [f'binary_delimiter_colon_{name}.csv' for name in subnames]
    input_data_path = './autoai/data/insurance_dataset/'

    input_node_names = subnames

    cos_endpoint = "https://s3.us.cloud-object-storage.appdomain.cloud"
    data_cos_path = "obm_data/"

    SPACE_ONLY = True

    BATCH_DEPLOYMENT_WITH_DA = True
    BATCH_DEPLOYMENT_WITH_CDA = True
    BATCH_DEPLOYMENT_WITH_CA_DA = True
    BATCH_DEPLOYMENT_WITH_CA_CA = True

    OPTIMIZER_NAME = "Binary Insurance dataset with colon delimeter OBM test sdk"

    target_space_id: str = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        prediction_type=PredictionType.BINARY,
        prediction_column='valid',
        positive_label=1,
        scoring=Metrics.RECALL_SCORE,
        drop_duplicates=False,
        include_only_estimators=[ClassificationAlgorithms.SnapSVM, ClassificationAlgorithms.EX_TREES]
    )

    def test_00b_prepare_COS_instance_and_connection(self):
        TestAutoAIRemote.connection_id, TestAutoAIRemote.bucket_name = create_connection_to_cos(
            wml_client=self.wml_client,
            cos_credentials=self.cos_credentials,
            cos_endpoint=self.cos_endpoint,
            bucket_name=self.bucket_name,
            save_data=True,
            data_path=self.input_data_path,
            data_filenames=self.input_data_filenames,
            data_cos_path=self.data_cos_path)

        self.assertIsInstance(self.connection_id, str)

    def test_01_create_multiple_data_connections__connections_created(self):
        TestAutoAIRemote.data_connections = []
        for file, node_name in zip(self.input_data_filenames, self.input_node_names):
            conn = DataConnection(
                data_join_node_name=node_name,
                connection_asset_id=self.connection_id,
                location=S3Location(
                    bucket=self.bucket_name,
                    path=os.path.join(self.data_cos_path, file)
                )
            )
            TestAutoAIRemote.data_connections.append(conn)

        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connections)
        self.assertEqual(len(TestAutoAIRemote.data_connections), len(self.input_data_filenames))
        self.assertIsNone(self.results_connection)

    def test_02_create_data_join_graph__graph_created(self):
        print("Defining DataJoinGraph...")
        data_join_graph = DataJoinGraph()
        data_join_graph.node(name="customer", csv_separator=":")
        data_join_graph.node(name="insurance", csv_separator=":", main=True)
        data_join_graph.node(name="claim", csv_separator=":")
        data_join_graph.edge(from_node="customer", to_node="insurance",
                             from_column=["customer_number"], to_column=["customer_id"])
        data_join_graph.edge(from_node="claim", to_node="insurance",
                             from_column=["claim_SN"], to_column=["claim_id"])

        TestAutoAIRemote.data_join_graph = data_join_graph

        print(f"data_join_graph: {data_join_graph}")

    def test_40b_deployment_input_files_setup(self):
        TestAutoAIRemote.input_data_filenames = self.deployment_payload_filenames


if __name__ == '__main__':
    unittest.main()
